// Copyright 2022 hjiayz
// 
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
//     http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


import {
    AST,
    RegExpParser,
    RegExpValidator,
    parseRegExpLiteral,
    validateRegExpLiteral,
    visitRegExpAST,
} from "regexpp";

import unicode from "./unicode";
import {caseFold,propertyValueAliases,propertyAliases} from "./unicode"

let fold : Map<number,number[]> = new Map();

for (let item of caseFold) {
    if (!fold.has(item[1])) {
        fold.set(item[1],[]);
    }
    fold.get(item[1]).push(item[0]);
}

type Global = {
    l2r: boolean,
}

export default class RegExp2Rust {
    ast: AST.RegExpLiteral
    pattern: AST.Pattern
    offset2id: Record<number, number>;
    group2id: Record<string, number>;
    groupenumstr: string;
    count: number;
    constructor(regex: RegExp) {
        this.ast = parseRegExpLiteral(regex);
        this.pattern = this.ast.pattern;
        [this.offset2id, this.group2id, this.count] = buildCapturingGroupIDS(this.ast);
        this.groupenumstr = buildGroupEnum(this.group2id);
        let flags = this.ast.flags;
        if (!flags.unicode) {
            throw "flags missing unicode(u)";
        }
        if (flags.hasIndices) {
            throw "flags hasIndices(d) not impl";
        }
    }
    build() {
        let f: string[] = [];
        let falts = newf(f);
        this.genAlts(this.pattern.alternatives, f, { l2r: true },falts,null);
        let iter = "";
        if (this.ast.flags.global) {
            iter = this.genIter();
        }
        let isGlobal = this.ast.flags.global;
        let isSticky = this.ast.flags.sticky;
        return `
        

        ${f.join("")}

        ///${this.pattern.raw}
        #[inline]
        pub fn exec(s:&str${(isSticky||isGlobal)?",mut lastIndex:usize":""})->[Option<[usize;2]>;${this.count + 1}]{
            ${(!(isGlobal||isSticky))?"let mut lastIndex = 0usize;":""}
            ${isSticky?``:`while lastIndex<=s.len() {`}
                let mut end = lastIndex;
                let mut matches : [Option<[usize;2]>;${this.count + 1}] = [None;${this.count + 1}];
                if ${fn(falts)}(s,&mut end,&mut matches) {
                    matches[0]=Some([lastIndex,end]);
                    return matches;
                }
            ${isSticky?``:`
                // https://tools.ietf.org/html/rfc3629
                const UTF8_CHAR_WIDTH: &[u8; 256] = &[
                    // 1  2  3  4  5  6  7  8  9  A  B  C  D  E  F
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 0
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 1
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 2
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 3
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 4
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 5
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 6
                    1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, // 7
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 8
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // 9
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // A
                    0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // B
                    0, 0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, // C
                    2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, // D
                    3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, // E
                    4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // F
                ];
                if lastIndex==s.len() {
                    lastIndex=usize::MAX;
                }
                else {
                    lastIndex=lastIndex + (UTF8_CHAR_WIDTH[s.as_bytes()[lastIndex] as usize] as usize);
                }
            }`}
            [None;${this.count + 1}]
        }
        #[derive(Copy,Clone,Debug)]
        pub enum Group{
${this.groupenumstr}
        }
        ${isGlobal?this.genIter():""}
        #[inline]
        pub fn matches(s:&str${(isSticky||isGlobal)?",lastIndex:usize":""})->[Option<&str>;${this.count + 1}]{
            exec(s${(isSticky||isGlobal)?",lastIndex":""}).map(|v|match v{
                Some([start,end])=>Some(&s[start..end]),
                None=>None,
            })
        }
        pub struct Matches<'a>(&'a str,[Option<[usize;2]>;${this.count + 1}]);
        pub fn typed_matches(s:&str${(isSticky||isGlobal)?",lastIndex:usize":""})->Matches<'_>{
            let matched = exec(s${(isSticky||isGlobal)?",lastIndex":""});
            Matches(s,matched)
        }
        impl<'a> Matches<'a>{
            pub fn get_str(&self,group:Group)->Option<&'a str>{
                let range = self.get_range(group)?;
                Some(&self.0[range[0]..range[1]])
            }
            pub fn get_range(&self,group:Group)->Option<[usize;2]>{
                self.1[group as usize]
            }
            pub fn get_str_id(&self,id:usize)->Option<&'a str>{
                let range = self.get_range_id(id)?;
                Some(&self.0[range[0]..range[1]])
            }
            pub fn get_range_id(&self,id:usize)->Option<[usize;2]>{
                if id>=self.1.len() {
                    return None;
                }
                self.1[id]
            }
        }
`
    }
    genIter(): string {
        return `
        pub struct Iter<'a>(&'a str,usize);
        impl<'a> Iter<'a> {
            #[inline]
            pub fn new(s:&'a str,lastIndex:usize)->Iter<'a>{
                Iter(s,lastIndex)
            }
        }
        impl core::iter::Iterator for Iter<'_> {
            type Item=[usize;2];
            #[inline]
            fn next(&mut self) -> Option<[usize;2]> {
                match exec(self.0,self.1)[0] {
                    Some([start,end])=>{
                        self.1=end;
                        Some([start,end])
                    }
                    None=>{
                        None
                    }
                }
            }
        }

        pub struct IterStr<'a>(&'a str,usize);
        impl<'a> IterStr<'a> {
            #[inline]
            pub fn new(s:&'a str,lastIndex:usize)->IterStr<'a>{
                IterStr(s,lastIndex)
            }
        }
        impl<'a> core::iter::Iterator for IterStr<'a> {
            type Item=&'a str;
            #[inline]
            fn next(&mut self) -> Option<&'a str> {
                match exec(self.0,self.1)[0] {
                    Some([start,end])=>{
                        self.1=end;
                        Some(&self.0[start..end])
                    }
                    None=>{
                        None
                    }
                }
            }
        }
        `
    }
    genAlts(alts: AST.Alternative[], f: string[], global: Global,id:number,next:number|void) {
        let s: string[] = [];
        let raws = [];
        for (let alt of alts) {
            raws.push(alt.raw);
            let fid = newf(f);
            this.genAlt(alt, f, global,fid,null);
            s.push(
`
            if ${fn(fid)}(s,offset,matches) {
                return ${nextstmt(next)};
            };
`);
        }
        let fname = fn(id);
        f[id]=`
        ///${raws.join("|")}
        #[inline]
        fn ${fname}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
${s.join("")}
            false
        }
`
    }
    genAlt(alt: AST.Alternative, f: string[], global: Global,id:number,next:number|void) {
        let eles = alt.elements;
        if (global.l2r==false) {
            eles = eles.reverse();
        }
        let fids = eles.map(()=>newf(f));
        for (let eleid in eles) {
            this.genEle(eles[eleid], f, global,fids[eleid],fids[parseInt(eleid)+1]);
        }

        let fname = fn(id);
        f[id]=`
        ///${alt.raw}
        #[inline]
        fn ${fname}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            let offset2 = *offset;
            let matches2 = *matches;
            if !${fn(fids[0])}(s,offset,matches) {
                *offset=offset2;
                *matches=matches2;
                return false;
            }
            ${nextstmt(next)}
        }
`
    }
    genEle(ele: AST.Element, f: string[], global: Global,id:number,next:number|void) {
        if (ele.type=="Assertion") {
            return this.genAssertion(ele,f,global,id,next);
        }
        if (ele.type=="Character") {
            return this.genCharacter(ele,f,global,id,next);
        }
        if (ele.type=="CharacterClass") {
            return this.genCharacterClass(ele,f,global,id,next);
        }
        if (ele.type=="Backreference") {
            return this.genBackreference(ele,f,global,id,next);
        }
        if (ele.type=="CapturingGroup") {
            return this.genCapturingGroup(ele,f,global,id,next);
        }
        if (ele.type=="CharacterSet") {
            return this.genCharacterSet(ele,f,global,id,next);
        }
        if (ele.type=="Group") {
            return this.genGroup(ele,f,global,id,next);
        }
        if (ele.type=="Quantifier") {
            return this.genQuantifier(ele,f,global,id,next);
        }
    }
    genGroup(ele: AST.Group, f: string[], global: Global,id:number,next:number|void) {
        let altsid = newf(f);
        this.genAlts(ele.alternatives,f,global,altsid,null);
        f[id]=`
        ///${ele.raw}
        #[inline]
        fn ${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            if ${fn(altsid)}(s,offset,matches) {
                return ${nextstmt(next)};
            }
            false
        }
        `
    }
    genQuantifier(ele: AST.Quantifier, f: string[], global: Global,id:number,next:number|void) {
        if (ele.greedy) {
            return this.genQuantifierGreedy(ele,f,global,id,next);
        }
        let max = ele.max;
        let min = ele.min;
        let checkmax = "";
        if (Number.POSITIVE_INFINITY!=max) {
            checkmax = `
                if count > ${max} {
                    *offset = offset2;
                    return false
                }`;
        }
        let fid = newf(f);
        this.genEle(ele.element,f,global,fid,null);

        let trynext = `
                        return true;
`;
        if (typeof next == "number") {
            trynext = `
                    if ${fn(next)}(s,offset,matches) {
                        return true;
                    }
`
        }

        f[id]=`
        ///${ele.raw}
        #[inline]
        fn ${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            let mut offset2 = *offset;
            let mut count : usize = 0;
            loop {
                if count >= ${min} {
                    ${trynext}
                }
                if !${fn(fid)}(s,offset,matches) {
                    *offset = offset2;
                    return false;
                }
                count+=1;
${checkmax}
            }

        }
        `
    }
    genQuantifierGreedy(ele: AST.Quantifier, f: string[], global: Global,id:number,next:number|void) {
        let max = ele.max;
        let min = ele.min;
        let checkmax = "";
        if (Number.POSITIVE_INFINITY!=max) {
            checkmax = `(n <= ${max}) &&`;
        }
        let fid = newf(f);
        this.genEle(ele.element,f,global,fid,null);

        let trynext = `
                *offset = offset2;
                *matches = matches2;
                return true;
`;
        if (typeof next == "number") {
            trynext = `
                if ${fn(next)}(s,&mut offset2,&mut matches2) {
                    *offset = offset2;
                    *matches = matches2;
                    return true;
                }
                else {
                    return false;
                }
`
        }

        f[id]=`
        ///${ele.raw}
        #[inline]
        fn ${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            fn search${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}],n:usize)->bool{
                let mut offset2 = *offset;
                let mut matches2 = *matches;
                if ${checkmax} ${fn(fid)}(s,&mut offset2,&mut matches2) {
                    if search${fn(id)}(s,&mut offset2,&mut matches2,n+1) {
                        *offset = offset2;
                        *matches = matches2;
                        return true;
                    }
                }
                if n < ${min} {
                    return false;
                }
${trynext}
            }
            search${fn(id)}(s,offset,matches,0)
        }
`
    }
    genCharacterSet(ele: AST.AnyCharacterSet | AST.EscapeCharacterSet | AST.UnicodePropertyCharacterSet, f: string[], global: Global,id:number,next:number|void) {
        return CharRanges.fromCharacterSet(ele,this.ast.flags.dotAll).gen(ele.raw,f,global,this.ast.flags.ignoreCase,this.count,id,next,true);
    }
    genCapturingGroup(ele: AST.CapturingGroup, f: string[], global: Global,id:number,next:number|void) {
        let groupid = this.offset2id[ele.start];
        let altsid = newf(f);
        this.genAlts(ele.alternatives,f,global,altsid,null);
        let range = global.l2r ? "start,end" : "end,start";
        f[id]=`
        ///${ele.raw}
        #[inline]
        fn ${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            let start = *offset;
            if ${fn(altsid)}(s,offset,matches) {
                let end = *offset;
                matches[${groupid}]=Some([${range}]);
                return ${nextstmt(next)};
            }
            false
        }
        `
    }
    genBackreference(ele: AST.Backreference, f: string[], global: Global,id:number,next:number|void) {
        let refid:number;
        if (typeof ele.ref == "string") {
            refid = this.group2id[ele.ref];
            if (typeof refid != "number") {
                throw "bug";
            }
        }
        else {
            refid = ele.ref;
        }
        let cmp: string;
        let setnewoffset: string;
        if (global.l2r) {
            cmp = "s[*offset..].starts_with(matched)";
            setnewoffset = "*offset+=len;";
        }
        else {
            cmp = "s[..*offset].ends_with(matched)";
            setnewoffset = "*offset-=len;";
        }
        f[id]=`
        ///${ele.raw}
        #[inline]
        fn ${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            let refid = ${refid};
            match matches[refid]{
                Some([start,end])=>{
                    let matched = &s[start..end];
                    let len = matched.len();
                    if ${cmp} {
                        ${setnewoffset}
                        true
                    }
                    else {
                        false
                    }
                }
                None=>false,
            }
        }
        `
    }
    genCharacterClass(ele: AST.CharacterClass, f: string[], global: Global,id:number,next:number|void) {
        return CharRanges.fromCharacterClass(ele,this.ast.flags.dotAll).gen(ele.raw,f,global,this.ast.flags.ignoreCase,this.count,id,next,true);
    }
    genCharacter(ele: AST.Character, f: string[], global: Global,id:number,next:number|void) {
        return CharRanges.fromCharacter(ele).gen(ele.raw,f,global,this.ast.flags.ignoreCase,this.count,id,next,true);
    }
    genAssertion(ele: AST.Assertion, f: string[], global: Global,id:number,next:number|void) {
        if (ele.kind=="start") {
            return this.genStartAssertion(ele,f,global,id,next);
        }
        if (ele.kind=="end") {
            return this.genEndAssertion(ele,f,global,id,next);
        }
        if (ele.kind=="lookahead") {
            return this.genLookahead(ele,f,global,id,next);
        }
        if (ele.kind=="lookbehind") {
            return this.genLookbehind(ele,f,global,id,next);
        }
        if (ele.kind=="word") {
            return this.genWordBoundary(ele,f,global,id,next);
        }
    }
    genWordBoundary(ele: AST.WordBoundaryAssertion, f: string[], global: Global,id:number,next:number|void) {
        let spaceRanges = CharRanges.whiteSpace();
        let spaceCheckPrevID = newf(f);
        spaceRanges.gen("spaceCheckPrev",f,Object.assign({},global,{l2r:false}),this.ast.flags.ignoreCase,this.count,spaceCheckPrevID,null,false);
        let spaceCheckNextID = newf(f);
        spaceRanges.gen("spaceCheckNext",f,Object.assign({},global,{l2r:true}),this.ast.flags.ignoreCase,this.count,spaceCheckNextID,null,false);

        f[id]=`
        ///${ele.raw}
        #[inline]
        fn ${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            let prevIsSpace = (*offset==0) || ${fn(spaceCheckPrevID)}(s,offset,matches);
            let nextIsSpace = (s.len()==*offset) || ${fn(spaceCheckNextID)}(s,offset,matches);
            if prevIsSpace ${ele.negate?"==":"!="} nextIsSpace {
                return ${nextstmt(next)};
            }
            false
        }
        `
    }
    genLookbehind(ele: AST.LookbehindAssertion, f: string[], global: Global,id:number,next:number|void) {
        let altsid = newf(f);
        this.genAlts(ele.alternatives,f,Object.assign({},global,{l2r:false}),altsid,null);
        f[id]=`
        ///${ele.raw}
        #[inline]
        fn ${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            let mut offset2 = *offset;
            let mut matches2 = *matches;
            if ${ele.negate?"!":""}${fn(altsid)}(s,&mut offset2,&mut matches2) {
                *matches = matches2;
                return ${nextstmt(next)};
            }
            false
        }
        `
    }
    genLookahead(ele: AST.LookaheadAssertion, f: string[], global: Global,id:number,next:number|void) {
        let altsid = newf(f);
        this.genAlts(ele.alternatives,f,Object.assign({},global,{l2r:true}),altsid,null);
        f[id]=`
        ///${ele.raw}
        #[inline]
        fn ${fn(id)}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            let mut offset2 = *offset;
            let mut matches2 = *matches;
            if ${ele.negate?"!":""}${fn(altsid)}(s,&mut offset2,&mut matches2) {
                *matches = matches2;
                return ${nextstmt(next)};
            }
            false
        }
        `
    }
    genStartAssertion(ele: AST.EdgeAssertion, f: string[], __: Global,id:number,next:number|void) {

        let multiline = "";
        if (this.ast.flags.multiline){
            let lineTerminator = CharRanges.lineTerminator();
            let lineTerminatorCheckPrevID = newf(f);
            lineTerminator.gen("lineTerminatorCheckPrev",f,Object.assign({},global,{l2r:false}),this.ast.flags.ignoreCase,this.count,lineTerminatorCheckPrevID,null,false);    
            multiline = ` || ${fn(lineTerminatorCheckPrevID)}(s,offset,matches) `;
        }
        let fname = fn(id);
        f[id]=
`
        ///${ele.raw}
        #[inline]
        fn ${fname}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            if *offset == 0 ${multiline} {
                ${nextstmt(next)}
            }
            else {
                false
            }
            
        }
`
    }
    genEndAssertion(ele: AST.EdgeAssertion, f: string[], __: Global,id:number,next:number|void) {
        let multiline = "";
        if (this.ast.flags.multiline){
            let lineTerminator = CharRanges.lineTerminator();
            let lineTerminatorCheckNextID = newf(f);
            lineTerminator.gen("lineTerminatorCheckNext",f,Object.assign({},global,{l2r:true}),this.ast.flags.ignoreCase,this.count,lineTerminatorCheckNextID,null,false);
            multiline = ` || ${fn(lineTerminatorCheckNextID)}(s,offset,matches) `;
        }
        let fname = fn(id);
        f[id]=
`
        ///${ele.raw}
        #[inline]
        fn ${fname}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${this.count + 1}])->bool{
            if *offset == s.len() ${multiline} {
                ${nextstmt(next)}
            }
            else {
                false
            }
        }
`
    }
}


function buildCapturingGroupIDS(tree: AST.Node): [Record<number, number>, Record<string, number>, number] {
    let list: number[] = [];
    let group2offset: Record<string, number> = {};
    let n = 0;
    visitRegExpAST(tree, {
        onCapturingGroupEnter: (node) => {
            list.push(node.start);
            if (node.name) {
                group2offset[node.name] = node.start;
            }
            n++;
        }
    });
    list.sort((a, b) => a - b);
    let offset2id: Record<number, number> = {};
    for (let id in list) {
        offset2id[list[id]] = parseInt(id) + 1;
    }
    let group2id: Record<string, number> = {};

    for (let key of Object.keys(group2offset)) {
        group2id[key] = offset2id[group2offset[key]];
    }
    return [offset2id, group2id, n];
}

function buildGroupEnum(group2id: Record<string, number>): string {
    let s: string[] = [];
    for (let name of Object.keys(group2id)) {
        s.push(`            ${name} = ${group2id[name]},`)
    }
    return s.join("\n")
}


class CharRanges{
    static MAX=0x10FFFF;
    static MIN=0;
    ranges:[number,number][]
    constructor(ranges:[number,number][]){
        this.ranges=ranges;
    }
    static empty():CharRanges{
        return new CharRanges([]);
    }
    static word():CharRanges{
        return new CharRanges([[0x30,0x39],[0x41,0x5A],[0x5f,0x5f],[0x61,0x7A]]);
    }
    static fromCharacterClass(ele:AST.CharacterClass,dotall:boolean):CharRanges{
        let base = this.empty();
        for(let ee of ele.elements) {
            if (ee.type=="Character"){
                base.union(CharRanges.fromCharacter(ee));
            } else if (ee.type == "CharacterSet") {
                base.union(CharRanges.fromCharacterSet(ee,dotall));
            } else if (ee.type == "CharacterClassRange") {
                base.union(CharRanges.fromCharacterClassRange(ee));
            }
        }
        if (ele.negate) {
            base.negate();
        }
        return base;
    }
    negate() {
        this.invert();
    }
    invert() {
        let ranges = this.ranges;
        let newRanges:[number,number][] = [];
        if (ranges.length==0) {
            newRanges.push([CharRanges.MIN,CharRanges.MAX]);
            this.ranges=newRanges;
            return ;
        }
        if (ranges[0][0]!=0) {
            newRanges.push([CharRanges.MIN,ranges[0][0]-1])
        }
        for(let i=0,j=1;j<ranges.length;i++,j++) {
            newRanges.push([ranges[i][1]+1,ranges[j][0]-1]);
        }
        let last = ranges[ranges.length-1];
        if (last[1]<CharRanges.MAX) {
            newRanges.push([last[1]+1, CharRanges.MAX]);
        }
        this.ranges=newRanges;
    }
    intersect(another:CharRanges) {
        let lr = this.ranges;
        let rr = another.ranges;
        let result:[number,number][] = [];
        for(let l=0,r=0;(l<lr.length) && (r<rr.length);){
            let lv = lr[l];
            let rv = rr[r];
            let start = Math.max(lv[0],rv[0]);
            let end:number ;
            if (lv[1]<rv[1]) {
                end = lv[1];
                l++;
            }
            else {
                end = rv[1];
                r++;
            }
            if (start > end) {
                continue;
            }
            let cur:[number,number] = [start, end];
            if (result.length==0) {
                result.push(cur);
                continue;
            }
            let prev = result[result.length-1];
            if (prev[1] < (cur[0]-1)) {
                result.push(cur);
                continue;
            }
            prev[1] = Math.max(prev[1], cur[1]);
        }
        this.ranges=result;
    }
    static fromCharacterClassRange(ee: AST.CharacterClassRange): CharRanges {
        return new CharRanges([[ee.min.value,ee.max.value]]);
    }
    union(another: CharRanges) {
        this.ranges=this.ranges.concat(another.ranges);
        this.reorganize();
    }
    static fromCharacterSet(ee:AST.AnyCharacterSet| AST.EscapeCharacterSet | AST.UnicodePropertyCharacterSet,dotall:boolean): CharRanges {
        let ranges:CharRanges;
        if (ee.kind=="digit") {
            ranges = new CharRanges([["0".codePointAt(0) as number,"9".charCodeAt(0) as number]]);
        } else if (ee.kind=="space") {
            ranges = CharRanges.whiteSpace();
            ranges.union(CharRanges.lineTerminator());
        } else if (ee.kind=="word") {
            ranges = CharRanges.word();
        } else if (ee.kind=="property"){
            ranges = CharRanges.property(ee.key,ee.value);
        } else if (ee.kind=="any") {
            if (dotall){
                ranges = CharRanges.empty();
            }
            else {
                ranges = CharRanges.lineTerminator();
            }
            ranges.negate();
            return ranges;
        }
        if (ee.negate) {
            ranges.negate();
        }
        return ranges;
    }
    static property(propertyName:string,propertyValue:string|void): CharRanges {
        let key:string;
        propertyName = propertyAliases.get(propertyName)||propertyName;
        if ( typeof propertyValue !== "string") {
            propertyAliases.get(propertyName)||propertyName
            key = `${propertyName}`
        }
        else {
            propertyValue = propertyValueAliases.get(propertyName).get(propertyValue)||propertyValue;
            key = `${propertyName}=${propertyValue}`
        }
        if (unicode[key]===undefined) {
            throw "is a bug";
        }
        return new CharRanges(unicode[key]);
        
    }
    static lineTerminator(): CharRanges {
        return new CharRanges([[0x000A,0x000A],[0x000D,0x000D],[0x2028,0x2029]])
    }
    static whiteSpace(): CharRanges {
        let ws = new CharRanges([[0x0009,0x0009],[0x000B,0x000C],[0xFEFF,0xFEFF]]);
        ws.union(CharRanges.property("General_Category","Space_Separator"))
        return ws;
    }
    static fromCharacter(ele:AST.Character):CharRanges{
        return new CharRanges([[ele.value,ele.value]]);
    }
    private reorganize(){       
        if (this.ranges.length==0) {
            return ;
        }
        this.ranges.sort((a,b)=>a[0]-b[0]);
        let newranges = [this.ranges[0]];
        for(let item of this.ranges ) {
            if (item[1]<item[0]) {
                throw new Error("bad input");
            }
            if ((newranges[newranges.length-1][1]+1)>=item[0]) {
                // cross
                newranges[newranges.length-1][1]=Math.max(newranges[newranges.length-1][1],item[1]);
            } else {
                newranges.push(item);
            }
        }
        this.ranges=newranges;
    }
    static ascii():CharRanges{
        return new CharRanges([[0,127]]);
    }
    static notAscii():CharRanges{
        let ascii = new CharRanges([[0,127]]);
        ascii.invert();
        return ascii;
    }
    isEmpty(){
        return this.ranges.length==0;
    }
    onlyOne(){
        return (this.ranges.length==1) && (this.ranges[0][0]==this.ranges[0][1]);
    }
    clone():CharRanges{
        let ranges:[number,number][] = [];
        for(let item of this.ranges) {
            ranges.push([item[0],item[1]]);
        }
        return new CharRanges(ranges);
    }
    onlyAscii():boolean{
        let cloned = this.clone();
        cloned.intersect(CharRanges.notAscii());
        return cloned.isEmpty();
    }
    unCaseFold(){
        const maybefold = new CharRanges(Array.from(fold.keys()).map(key=>[key,key]));
        maybefold.reorganize();
        maybefold.intersect(this);
        let folded = [];
        for(let [start,end] of maybefold.ranges ) {
            for(let item=start;item<=end ;item++) {
                let vals = fold.get(item);
                for (let val of vals){
                    folded.push([val,val]);
                }
            }
        }
        let ranges = new CharRanges(folded);
        ranges.reorganize();
        this.union(ranges);
    }
    gen(raw:string,f: string[], global: Global,ignoreCase:boolean,count:number,id:number,next:number|void,changeoffset:boolean){
        if (ignoreCase) {
            this.unCaseFold();
        }
        if (this.isEmpty()) {
            return this.genEmpty(raw,f,global,count,id,next,changeoffset);
        }
        if (this.onlyOne()) {
            return this.genOne(raw,f,global,count,id,next,changeoffset);
        }
        if (this.onlyAscii()) {
            return this.genAscii(raw,f,global,count,id,next,changeoffset);
        }
        return this.genUnicode(raw,f,global,count,id,next,changeoffset);
    }
    genOne(raw:string,f: string[], global: Global, count: number,id:number,next:number|void,changeoffset:boolean) {
        
        let fname = fn(id);
        let cmp = global.l2r ? (
`
            if s[*offset..].starts_with(ch) {
                ${changeoffset?"*offset = *offset + len;":""}
                ${nextstmt(next)}
            }
            else {
                false
            }
`
        ):(
            `
            if s[0..*offset].ends_with(ch) {
                ${changeoffset?"*offset = *offset - len;":""}
                ${nextstmt(next)}
            }
            else {
                false
            }
`
        );
        f[id]=
`
        ///${raw}
        #[inline]
        fn ${fname}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${count + 1}])->bool{
            const ch : &'static str = "\\u{${this.ranges[0][0].toString(16)}}";
            const len : usize = ch.len();
            ${cmp}
        }
`  
    }
    genEmpty(raw:string,f: string[], global: Global, count: number,id:number,next:number|void,changeoffset:boolean) {
        let fname = fn(id);
        f[id]=
`
        ///${raw}
        #[inline]
        fn ${fname}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${count + 1}])->bool{
            false
        }
`     
    }
    genAscii(raw:string,f: string[], global: Global, count: number,id:number,next:number|void,changeoffset:boolean) {
        
        let fname = fn(id);
        let okCase = this.genCase();
        let cmp = global.l2r ? (
`
            if *offset >= s.len() {
                return false
            }
            match b[*offset] {
                ${okCase}=>{
                    ${changeoffset?"*offset = *offset + 1;":""}
                    ${nextstmt(next)}
                }
                _ => {
                    false
                }
            }
`
        ):(
            `
            if *offset<1 {
                return false;
            }
            match b[*offset-1] {
                ${okCase}=>{
                    ${changeoffset?"*offset = *offset - 1;":""}
                    ${nextstmt(next)}
                }
                _ => {
                    false
                }
            }
`
        );
        f[id]=
`
        ///${raw}
        #[inline]
        fn ${fname}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${count + 1}])->bool{
            let b = s.as_bytes();
            ${cmp}
        }
`
    }
    genUnicode(raw:string,f: string[], global: Global, count: number,id:number,next:number|void,changeoffset:boolean) {
        let fname = fn(id);
        let okCase = this.genCase();
        let cmp = global.l2r ? (
`
            if *offset >= s.len() {
                return false
            }
            let ch = s[*offset..].chars().next().unwrap();
            match ch as u32 {
                ${okCase}=>{
                    ${changeoffset?"*offset = *offset + ch.len_utf8();":""}
                    ${nextstmt(next)}
                }
                _ => {
                    false
                }
            }
`
        ):(
            `
            if *offset<1 {
                return false;
            }
            let ch = s[0..*offset].chars().next_back().unwrap();
            match ch as u32 {
                ${okCase}=>{
                    ${changeoffset?"*offset = *offset - ch.len_utf8();":""}
                    ${nextstmt(next)}
                }
                _ => {
                    false
                }
            }
`
        );
        f[id]=
`
        ///${raw}
        #[inline]
        fn ${fname}(s:&str,offset:&mut usize,matches:&mut [Option<[usize;2]>;${count + 1}])->bool{
            ${cmp}
        }
`
    }
    genCase():string{
        let s = [];
        this.ranges.forEach(([start,end])=>{
            s.push(`0x${start.toString(16)}..=0x${end.toString(16)}`)
        })
        return s.join("|\n                ");
    }
}

function fn(id:number):string{
    return `f${id}`;
}
function newf(f:string[]):number{
    f.push("");
    return f.length-1;
}
function nextstmt(next:number|void):string{
    return (typeof next == "number") ? `${fn(next)}(s,offset,matches)` : "true";
}